from datetime import datetime, timedelta
from typing import ClassVar

from pydantic import field_serializer, field_validator, TypeAdapter

from liti.core.model.v1.operation.data.base import EntityKind, Operation
from liti.core.model.v1.schema import ColumnName, ConstraintName, ForeignKey, Identifier, IntervalLiteral, PrimaryKey, \
    QualifiedName, RoundingMode, Schema, StorageBilling, Table


class CreateSchema(Operation):
    """ Semantics: `CREATE SCHEMA` """

    schema_object: Schema

    KIND: ClassVar[str] = 'create_schema'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class DropSchema(Operation):
    """ Semantics: `DROP SCHEMA`

    The down migration for drop schema does NOT attempt to recreate anything within the schema. It only recreates the
    schema itself with the same properties it had when dropped. This behavior may change in the future.
    """

    schema_name: QualifiedName

    KIND: ClassVar[str] = 'drop_schema'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class SetDefaultTableExpiration(Operation):
    schema_name: QualifiedName
    expiration: timedelta | None = None

    KIND: ClassVar[str] = 'set_default_table_expiration'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}

    @field_serializer('expiration')
    @classmethod
    def serialize_timedelta(cls, value: timedelta | None) -> str | None:
        return value and TypeAdapter(timedelta).dump_python(value, mode='json')


class SetDefaultPartitionExpiration(Operation):
    schema_name: QualifiedName
    expiration: timedelta | None = None

    KIND: ClassVar[str] = 'set_default_partition_expiration'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}

    @field_serializer('expiration')
    @classmethod
    def serialize_timedelta(cls, value: timedelta | None) -> str | None:
        return value and TypeAdapter(timedelta).dump_python(value, mode='json')


class SetDefaultKmsKeyName(Operation):
    schema_name: QualifiedName
    key_name: str | None = None

    KIND: ClassVar[str] = 'set_default_kms_key_name'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class SetFailoverReservation(Operation):
    schema_name: QualifiedName
    reservation: str | None = None

    KIND: ClassVar[str] = 'set_failover_reservation'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class SetCaseSensitive(Operation):
    schema_name: QualifiedName
    case_sensitive: bool

    KIND: ClassVar[str] = 'set_case_sensitive'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class SetIsPrimaryReplica(Operation):
    schema_name: QualifiedName
    is_primary: bool

    KIND: ClassVar[str] = 'set_is_primary_replica'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class SetPrimaryReplica(Operation):
    schema_name: QualifiedName
    replica: str | None = None

    KIND: ClassVar[str] = 'set_primary_replica'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class SetMaxTimeTravel(Operation):
    schema_name: QualifiedName
    duration: timedelta | None = None

    KIND: ClassVar[str] = 'set_max_time_travel'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}

    @field_serializer('duration')
    @classmethod
    def serialize_timedelta(cls, value: timedelta | None) -> str | None:
        return value and TypeAdapter(timedelta).dump_python(value, mode='json')


class SetStorageBilling(Operation):
    schema_name: QualifiedName
    storage_billing: StorageBilling

    KIND: ClassVar[str] = 'set_storage_billing'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA'}


class CreateTable(Operation):
    """ Semantics: `CREATE TABLE` """

    table: Table

    KIND: ClassVar[str] = 'create_table'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class DropTable(Operation):
    """ Semantics: `DROP TABLE` """

    table_name: QualifiedName

    KIND: ClassVar[str] = 'drop_table'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class RenameTable(Operation):
    from_name: QualifiedName
    to_name: Identifier

    KIND: ClassVar[str] = 'rename_table'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetPrimaryKey(Operation):
    table_name: QualifiedName
    primary_key: PrimaryKey | None = None

    KIND: ClassVar[str] = 'set_primary_key'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class AddForeignKey(Operation):
    table_name: QualifiedName
    foreign_key: ForeignKey

    KIND: ClassVar[str] = 'add_foreign_key'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class DropConstraint(Operation):
    table_name: QualifiedName
    constraint_name: ConstraintName

    KIND: ClassVar[str] = 'drop_constraint'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetPartitionExpiration(Operation):
    table_name: QualifiedName
    expiration: timedelta | None = None

    KIND: ClassVar[str] = 'set_partition_expiration'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}

    @field_serializer('expiration')
    @classmethod
    def serialize_timedelta(cls, value: timedelta | None) -> str | None:
        return value and TypeAdapter(timedelta).dump_python(value, mode='json')


class SetRequirePartitionFilter(Operation):
    table_name: QualifiedName
    require_filter: bool

    KIND: ClassVar[str] = 'set_require_partition_filter'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetClustering(Operation):
    table_name: QualifiedName
    column_names: list[ColumnName] | None = None

    KIND: ClassVar[str] = 'set_clustering'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}

    @field_validator('column_names', mode='before')
    @classmethod
    def validate_column_names(cls, value: list[ColumnName] | None) -> list[ColumnName] | None:
        if value:
            return value
        else:
            return None


class SetFriendlyName(Operation):
    entity_name: QualifiedName
    friendly_name: str | None = None

    KIND: ClassVar[str] = 'set_friendly_name'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA', 'TABLE', 'VIEW', 'MATERIALIZED_VIEW'}


class SetDescription(Operation):
    entity_name: QualifiedName
    description: str | None = None

    KIND: ClassVar[str] = 'set_description'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA', 'TABLE', 'VIEW', 'MATERIALIZED_VIEW'}


class SetLabels(Operation):
    entity_name: QualifiedName
    labels: dict[str, str] | None = None

    KIND: ClassVar[str] = 'set_labels'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA', 'TABLE', 'VIEW', 'MATERIALIZED_VIEW'}


class SetTags(Operation):
    entity_name: QualifiedName
    tags: dict[str, str] | None = None

    KIND: ClassVar[str] = 'set_tags'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA', 'TABLE', 'VIEW', 'MATERIALIZED_VIEW'}


class SetExpirationTimestamp(Operation):
    entity_name: QualifiedName
    expiration_timestamp: datetime | None = None

    KIND: ClassVar[str] = 'set_expiration_timestamp'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE', 'VIEW', 'MATERIALIZED_VIEW'}


class SetDefaultRoundingMode(Operation):
    entity_name: QualifiedName
    rounding_mode: RoundingMode | None = None

    KIND: ClassVar[str] = 'set_default_rounding_mode'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'SCHEMA', 'TABLE'}


class SetMaxStaleness(Operation):
    entity_name: QualifiedName
    max_staleness: IntervalLiteral | None = None

    KIND: ClassVar[str] = 'set_max_staleness'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE', 'MATERIALIZED_VIEW'}


class SetEnableChangeHistory(Operation):
    table_name: QualifiedName
    enabled: bool

    KIND: ClassVar[str] = 'set_enable_change_history'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetEnableFineGrainedMutations(Operation):
    table_name: QualifiedName
    enabled: bool

    KIND: ClassVar[str] = 'set_enable_fine_grained_mutations'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetKmsKeyName(Operation):
    table_name: QualifiedName
    key_name: str | None = None

    KIND: ClassVar[str] = 'set_kms_key_name'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}



# noinspection PyUnresolvedReferences
from . import datatype, schema
# noinspection PyUnresolvedReferences
from .operation.data import subclasses
# noinspection PyUnresolvedReferences
from .operation.ops import subclasses



from typing import ClassVar

from liti.core.model.v1.operation.data.base import EntityKind, Operation
from liti.core.model.v1.schema import MaterializedView, QualifiedName, View


class CreateView(Operation):
    """ Semantics: `CREATE OR REPLACE VIEW` """

    view: View

    KIND: ClassVar[str] = 'create_view'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'VIEW'}


class DropView(Operation):
    """ Semantics: `DROP VIEW` """

    view_name: QualifiedName

    KIND: ClassVar[str] = 'drop_view'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'VIEW'}


class CreateMaterializedView(Operation):
    """ Semantics: `CREATE OR REPLACE MATERIALIZED VIEW` """

    materialized_view: MaterializedView

    KIND: ClassVar[str] = 'create_materialized_view'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'MATERIALIZED_VIEW'}


class DropMaterializedView(Operation):
    """ Semantics: `DROP MATERIALIZED VIEW` """

    materialized_view_name: QualifiedName

    KIND: ClassVar[str] = 'drop_materialized_view'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'MATERIALIZED_VIEW'}



from liti.core.context import Context

from liti.core.model.v1.datatype import Array, BigNumeric, Bytes, Float, Int, Numeric, String
from liti.core.model.v1.schema import MaterializedView, Partitioning, Schema, Table, View


class Defaulter:
    """ Observer interface for backends to implement to define defaults

    Default methods update None values to their defaults.
    """

    def int_defaults(self, node: Int, context: Context):
        pass

    def float_defaults(self, node: Float, context: Context):
        pass

    def numeric_defaults(self, node: Numeric, context: Context):
        pass

    def big_numeric_defaults(self, node: BigNumeric, context: Context):
        pass

    def partitioning_defaults(self, node: Partitioning, context: Context):
        pass

    def table_defaults(self, node: Table, context: Context):
        pass

    def view_defaults(self, node: View, context: Context):
        if node.select_sql is None and node.select_file is not None:
            with open(context.target_dir / node.select_file) as f:
                node.select_sql = f.read()

    def materialized_view_defaults(self, node: MaterializedView, context: Context):
        if node.select_sql is None and node.select_file is not None:
            with open(context.target_dir / node.select_file) as f:
                node.select_sql = f.read()


class Validator:
    """ Observer interface for backends to implement to validate the model

    Validation methods fix invalid values and raise if still invalid.
    """

    def validate_schema(self, node: Schema, context: Context):
        pass

    def validate_int(self, node: Int, context: Context):
        pass

    def validate_float(self, node: Float, context: Context):
        pass

    def validate_numeric(self, node: Numeric, context: Context):
        pass

    def validate_big_numeric(self, node: BigNumeric, context: Context):
        pass

    def validate_string(self, node: String, context: Context):
        pass

    def validate_bytes(self, node: Bytes, context: Context):
        pass

    def validate_array(self, node: Array, context: Context):
        pass

    def validate_partitioning(self, node: Partitioning, context: Context):
        pass

    def validate_view(self, node: View, context: Context):
        if not node.select_sql:
            raise ValueError(f'View {node.name} has no select SQL')

    def validate_materialized_view(self, node: MaterializedView, context: Context):
        if not node.select_sql:
            raise ValueError(f'Materialized view {node.name} has no select SQL')



from typing import Any, ClassVar, Literal

from liti.core.base import LitiModel
from liti.core.reflect import recursive_subclasses

EntityKind = Literal['SCHEMA', 'TABLE', 'VIEW', 'MATERIALIZED_VIEW']


class Operation(LitiModel):
    KIND: ClassVar[str]

    @classmethod
    def by_kind(cls, kind: str) -> type['Operation']:
        # ensure Operation subclasses are imported first
        # noinspection PyUnresolvedReferences
        import liti.core.model.v1.operation.data.subclasses

        return {
            subclass.KIND: subclass
            for subclass in recursive_subclasses(Operation)
        }[kind]

    def to_op_data(self, format: Literal['json', 'yaml']) -> dict[str, Any]:
        data = self.model_dump(
            mode='json' if format == 'json' else 'python',
            exclude_none=True,
        )

        return {
            'kind': self.KIND,
            'data': data,
        }

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return set()



from datetime import datetime, timedelta

from liti.core.backend.base import CreateRelation, DbBackend, MetaBackend
from liti.core.model.v1.datatype import Datatype
from liti.core.model.v1.operation.data.base import Operation
from liti.core.model.v1.operation.data.table import CreateSchema, CreateTable
from liti.core.model.v1.operation.data.view import CreateMaterializedView, CreateView
from liti.core.model.v1.schema import Column, ColumnName, ConstraintName, DatabaseName, ForeignKey, Identifier, \
    IntervalLiteral, MaterializedView, PrimaryKey, QualifiedName, RoundingMode, Schema, SchemaName, StorageBilling, \
    Table, View


class MemoryDbBackend(DbBackend):
    def __init__(self):
        self.schemas: dict[QualifiedName, Schema] = {}
        self.tables: dict[QualifiedName, Table] = {}
        self.views: dict[QualifiedName, View] = {}
        self.materialized_views: dict[QualifiedName, MaterializedView] = {}

    def scan_schema(self, database: DatabaseName, schema: SchemaName) -> list[Operation]:
        schema = self.get_schema(QualifiedName(database=database, schema_name=schema))

        if schema:
            create_schema = [CreateSchema(schema_object=schema)]
        else:
            create_schema = []

        tables = [
            CreateTable(table=table)
            for name, table in self.tables.items()
            if name.database == database and name.schema_name == schema
        ]

        views = [
            CreateView(view=view)
            for name, view in self.views.items()
            if name.database == database and name.schema_name == schema
        ]

        materialized_views = [
            CreateMaterializedView(materialized_view=materialized_view)
            for name, materialized_view in self.materialized_views.items()
            if name.database == database and name.schema_name == schema
        ]

        return create_schema + tables + materialized_views + views

    def scan_relation(self, name: QualifiedName) -> CreateRelation | None:
        if name in self.tables:
            return CreateTable(table=self.tables[name])
        if name in self.views:
            return CreateView(view=self.views[name])
        if name in self.materialized_views:
            return CreateMaterializedView(materialized_view=self.materialized_views[name])
        else:
            return None

    def get_schema(self, name: QualifiedName) -> Schema | None:
        return self.schemas.get(name)

    def create_schema(self, schema: Schema):
        if schema.name in self.schemas:
            raise ValueError(f'Schema {schema.name} already exists')

        self.schemas[schema.name] = schema.model_copy(deep=True)

    def drop_schema(self, name: QualifiedName):
        if name not in self.schemas:
            raise ValueError(f'Schema {name} does not exist')

        del self.schemas[name]

    def set_default_table_expiration(self, schema_name: QualifiedName, expiration: timedelta | None):
        self.schemas[schema_name].default_table_expiration = expiration

    def set_default_partition_expiration(self, schema_name: QualifiedName, expiration: timedelta | None):
        self.schemas[schema_name].default_partition_expiration = expiration

    def set_default_kms_key_name(self, schema_name: QualifiedName, key_name: str | None):
        self.schemas[schema_name].default_kms_key_name = key_name

    def set_failover_reservation(self, schema_name: QualifiedName, reservation: str | None):
        self.schemas[schema_name].failover_reservation = reservation

    def set_case_sensitive(self, schema_name: QualifiedName, case_sensitive: bool):
        self.schemas[schema_name].is_case_sensitive = case_sensitive

    def set_is_primary_replica(self, schema_name: QualifiedName, is_primary: bool):
        self.schemas[schema_name].is_primary_replica = is_primary

    def set_primary_replica(self, schema_name: QualifiedName, replica: str | None):
        self.schemas[schema_name].primary_replica = replica

    def set_max_time_travel(self, schema_name: QualifiedName, duration: timedelta | None):
        self.schemas[schema_name].max_time_travel = duration

    def set_storage_billing(self, schema_name: QualifiedName, storage_billing: StorageBilling):
        self.schemas[schema_name].storage_billing = storage_billing

    def get_table(self, name: QualifiedName) -> Table | None:
        return self.tables.get(name)

    def create_table(self, table: Table):
        if table.name in self.tables:
            raise ValueError(f'Table {table.name} already exists')

        self.tables[table.name] = table.model_copy(deep=True)

    def drop_table(self, name: QualifiedName):
        if name not in self.tables:
            raise ValueError(f'Table {name} does not exist')

        del self.tables[name]

    def rename_table(self, from_name: QualifiedName, to_name: Identifier):
        self.tables[from_name.with_name(to_name)] = self.tables.pop(from_name)

    def set_primary_key(self, table_name: QualifiedName, primary_key: PrimaryKey | None):
        self.tables[table_name].primary_key = primary_key

    def add_foreign_key(self, table_name: QualifiedName, foreign_key: ForeignKey):
        self.tables[table_name].add_foreign_key(foreign_key)

    def drop_constraint(self, table_name: QualifiedName, constraint_name: ConstraintName):
        self.tables[table_name].drop_constraint(constraint_name)

    def set_partition_expiration(self, table_name: QualifiedName, expiration: timedelta | None):
        self.tables[table_name].partitioning.expiration = expiration

    def set_require_partition_filter(self, table_name: QualifiedName, require_filter: bool):
        self.tables[table_name].partitioning.require_filter = require_filter

    def set_clustering(self, table_name: QualifiedName, column_names: list[ColumnName] | None):
        self.tables[table_name].clustering = column_names

    def set_friendly_name(self, entity_name: QualifiedName, friendly_name: str | None):
        self.get_entity(entity_name).friendly_name = friendly_name

    def set_description(self, entity_name: QualifiedName, description: str | None):
        self.get_entity(entity_name).description = description

    def set_labels(self, entity_name: QualifiedName, labels: dict[str, str] | None):
        self.get_entity(entity_name).labels = labels

    def set_tags(self, entity_name: QualifiedName, tags: dict[str, str] | None):
        self.get_entity(entity_name).tags = tags

    def set_expiration_timestamp(self, entity_name: QualifiedName, expiration_timestamp: datetime | None):
        self.get_entity(entity_name).expiration_timestamp = expiration_timestamp

    def set_default_rounding_mode(self, entity_name: QualifiedName, rounding_mode: RoundingMode | None):
        self.get_entity(entity_name).default_rounding_mode = rounding_mode

    def set_max_staleness(self, entity_name: QualifiedName, max_staleness: IntervalLiteral | None):
        self.get_entity(entity_name).max_staleness = max_staleness

    def set_enable_change_history(self, table_name: QualifiedName, enabled: bool):
        self.tables[table_name].enable_change_history = enabled

    def set_enable_fine_grained_mutations(self, table_name: QualifiedName, enabled: bool):
        self.tables[table_name].enable_fine_grained_mutations = enabled

    def set_kms_key_name(self, table_name: QualifiedName, key_name: str | None):
        self.tables[table_name].kms_key_name = key_name

    def add_column(self, table_name: QualifiedName, column: Column):
        self.get_table(table_name).columns.append(column.model_copy(deep=True))

    def drop_column(self, table_name: QualifiedName, column_name: ColumnName):
        table = self.get_table(table_name)
        table.columns = [col for col in table.columns if col.name != column_name]

    def rename_column(self, table_name: QualifiedName, from_name: ColumnName, to_name: ColumnName):
        table = self.get_table(table_name)
        table.columns = [col if col.name != from_name else col.with_name(to_name) for col in table.columns]

    def set_column_datatype(self, table_name: QualifiedName, column_name: ColumnName, from_datatype: Datatype, to_datatype: Datatype):
        table = self.get_table(table_name)
        column = table.column_map[column_name]
        column.datatype = to_datatype

    def set_column_nullable(self, table_name: QualifiedName, column_name: ColumnName, nullable: bool):
        table = self.get_table(table_name)
        column = table.column_map[column_name]
        column.nullable = nullable

    def set_column_description(self, table_name: QualifiedName, column_name: ColumnName, description: str | None):
        table = self.get_table(table_name)
        column = table.column_map[column_name]
        column.description = description

    def set_column_rounding_mode(
        self,
        table_name: QualifiedName,
        column_name: ColumnName,
        rounding_mode: RoundingMode | None,
    ):
        table = self.get_table(table_name)
        column = table.column_map[column_name]
        column.rounding_mode = rounding_mode

    def get_view(self, name: QualifiedName) -> View | None:
        return self.views.get(name)

    def create_view(self, view: View):
        self.views[view.name] = view.model_copy(deep=True)

    def drop_view(self, name: QualifiedName):
        if name not in self.views:
            raise ValueError(f'View {name} does not exist')

        del self.views[name]

    def get_materialized_view(self, name: QualifiedName) -> MaterializedView | None:
        return self.materialized_views.get(name)

    def create_materialized_view(self, materialized_view: MaterializedView):
        self.materialized_views[materialized_view.name] = materialized_view.model_copy(deep=True)

    def drop_materialized_view(self, name: QualifiedName):
        if name not in self.materialized_views:
            raise ValueError(f'MaterializedView {name} does not exist')

        del self.materialized_views[name]


class MemoryMetaBackend(MetaBackend):
    def __init__(self, applied_operations: list[Operation] | None = None):
        self.applied_operations = applied_operations or []

    def get_applied_operations(self) -> list[Operation]:
        return self.applied_operations

    def apply_operation(self, operation: Operation):
        self.applied_operations.append(operation)

    def unapply_operation(self, operation: Operation):
        most_recent = self.applied_operations.pop()
        assert operation == most_recent, 'Expected the operation to be the most recent one'



from typing import ClassVar

from liti.core.model.v1.operation.data.base import Operation
from liti.core.model.v1.schema import QualifiedName


class ExecuteSql(Operation):
    """ Run arbitrary SQL

    Only use this as a last resort.
    Limber Timber is not designed to be used this way.
    The primary use case is DML migrations what cannot be described by liti types.
    All fields are relative paths from the target directory to a SQL file.
    The paths are serialized to the metadata, not the SQL contents.
    It is highly recommended not to change the content of the SQL files, Limber Timber will not detect the changes.

    :param up: path to a SQL script to execute the up migration, must be an atomic operation
    :param down: path to a SQL script to execute the down migration, must be an atomic operation
    :param is_up: path to a SQL file with a boolean value query
        the query must return TRUE if the up migration has been applied
        the query must return FALSE if the up migration has not been applied
        TRUE and FALSE behave as if the query returned that value
    :param is_down: path to a SQL file with a boolean value query
        the query must return TRUE if the down migration has been applied
        the query must return FALSE if the down migration has not been applied
        TRUE and FALSE behave as if the query returned that value
    :param entity_names: mapping from python string format keys to fully qualified names
        SQL can be written as a python format string with named parameters
        the parameters will be replaced with the fully qualified names using str.format
        this field is provided to allow for templating tables, views, etc.
    """

    up: str
    down: str
    is_up: str | bool = False
    is_down: str | bool = False
    entity_names: dict[str, QualifiedName] | None = None

    KIND: ClassVar[str] = 'execute_sql'



from math import isclose
from types import UnionType
from typing import Any, Callable, Generator, get_args, get_origin

from pydantic import BaseModel, TypeAdapter

from liti.core.reflect import recursive_subclasses


def is_match(match: Any, value: Any) -> bool:
    # circular import
    from liti.core.model.v1.schema import ValidatedString

    if isinstance(match, dict) and isinstance(value, list | tuple | set):
        # skip checking collection items
        return True
    elif isinstance(match, str) and isinstance(value, dict):
        return match in value
    elif isinstance(match, dict) and isinstance(value, dict):
        return all(mk in value and is_match(mv, value[mk]) for mk, mv in match.items())
    elif isinstance(match, dict) and isinstance(value, LitiModel):
        # dig deeper into the model
        return all(is_match(inner, getattr(value, field)) for field, inner in match.items())
    elif isinstance(match, float) and isinstance(value, float):
        return isclose(match, value)
    elif isinstance(value, ValidatedString):
        # avoids having to specify '.string' in templates
        return match == value.string
    else:
        # match must be on the left hand side so STAR uses its __eq__ implementation
        return match == value


class Star:
    """ Star is used to match everything """

    def __eq__(self, other):
        return True

    def __getitem__(self, item):
        return self

    def get(self, *args, **kwargs):
        return self

    def items(self):
        return iter(())


STAR = Star()


class LitiModel(BaseModel):
    """ Base class for all Liti model classes """

    @classmethod
    def by_name(cls, name: str) -> type['LitiModel']:
        # ensure LitiModel subclasses are imported first
        # noinspection PyUnresolvedReferences
        import liti.core.model.v1.subclasses

        return {
            subclass.__name__: subclass
            for subclass in recursive_subclasses(LitiModel)
        }[name]

    def get_roots(self, root: type['LitiModel'], full_match: Any) -> Generator[tuple['LitiModel', Any], None, None]:
        """ Yields all the nodes of type `root` that match `full_match`

        Also yields the remaining field matches associated with each root in case the template path traverses through
        collection nodes. Those field matches can be used to check each item in the collections.
        """

        # this can be a bit duplicative since each call to `is_match` is recursive,
        # but performance here is not a concern, this is the easiest way to implement
        # the matching logic
        if not is_match(full_match, self):
            return

        if isinstance(self, root):
            # TODO: this does not handle recursive datatypes, update the logic if we ever have recursive LitiModels
            yield self, full_match
        else:
            for field_name in self.__pydantic_fields__.keys():
                field = getattr(self, field_name)

                if isinstance(field, list | tuple | set):
                    for item in field:
                        if isinstance(item, LitiModel):
                            yield from item.get_roots(root, full_match.get(field_name, STAR))
                elif isinstance(field, dict) and field_name in full_match:
                    match = full_match[field_name]

                    if match is STAR or isinstance(match, dict):
                        for key, item in field.items():
                            if key in match and isinstance(item, LitiModel):
                                yield from item.get_roots(root, match[key])
                    else:
                        item = field.get(match)

                        if isinstance(item, LitiModel):
                            yield from item.get_roots(root, STAR)
                elif isinstance(field, LitiModel):
                    yield from field.get_roots(root, full_match.get(field_name, STAR))

    def get_update_fns(self, path: list[str], matches: list[Any]) -> Generator[Callable[[Any], None], None, None]:
        """ Yields functions to replace selected fields with a provided value

        :param path: a list of field names to recursively traverse through to find the fields to update
        :param matches: a list of either a dict structure of values to compare to the respective fields (functions are
            yielded on equivalence of all fields), or Star to always yield a function, all items in the list must match
        """

        # circular import
        from liti.core.model.v1.schema import ValidatedString

        if not path:
            return

        field_name, *tail = path

        if not hasattr(self, field_name):
            return

        field = getattr(self, field_name)
        field_matches = [m.get(field_name, STAR) for m in matches]

        # stop if any sibling fields do not match
        if not all(
            hasattr(self, f) and is_match(inner, getattr(self, f))
            for m in matches
            for f, inner in m.items()
            if f != field_name
        ):
            return

        if tail:
            # if there are more segments, dig deeper into the model
            if isinstance(field, tuple | list | set):
                # yield for each item that matches in the collection
                for item in field:
                    if isinstance(item, LitiModel):
                        yield from item.get_update_fns(tail, field_matches)
            elif isinstance(field, dict):
                key, *dict_tail = tail
                item = field.get(key)

                if dict_tail:
                    # dig deeper
                    if isinstance(item, LitiModel):
                        yield from item.get_update_fns(dict_tail, field_matches)
                # yield the leaf dict value if it matches
                elif all(is_match(fm, item) for fm in field_matches):
                    yield lambda value: field.update({key: value})
            elif isinstance(field, LitiModel):
                yield from field.get_update_fns(tail, field_matches)
        # yield the leaf field if it matches
        elif all(is_match(fm, field) for fm in field_matches):
            field_type = self.__pydantic_fields__[field_name].annotation
            subclass = extract_subclass(field_type, ValidatedString)

            if subclass is not None:
                # avoids having to specify '.string' in templates when the field is a ValidatedString
                yield lambda value: setattr(self, field_name, subclass(value))
            else:
                yield lambda value: setattr(self, field_name, TypeAdapter(field_type).validate_python(value))


def extract_subclass(ty: type, parent: type) -> type | None:
    """ Returns the first encountered subclass of parent if any exist, supports plain subclasses and union types

    This is needed to support instantiation of the proper ValidatedString subtypes based on field type information.
    We cannot instantiate a Union[Identifier, None], but we can instantiate an Identifier and assign it to a field of
    type Union[Identifier, None].
    """

    if issubclass(ty, parent):
        return ty

    if get_origin(ty) is UnionType:
        for t in get_args(ty):
            if issubclass(t, parent):
                return t

    return None



import json
from pathlib import Path

import yaml


def parse_json_or_yaml_file(path: Path) -> list | dict:
    with open(path) as f:
        content = f.read()

    suffix = path.suffix.lower()

    if suffix == '.json':
        return json.loads(content)
    elif suffix in ('.yaml', '.yml'):
        return yaml.safe_load(content)
    else:
        raise ValueError(f'Unexpected file extension: "{path}"')


def get_manifest_path(target_dir: Path) -> Path:
    filenames = ('manifest.json', 'manifest.yaml', 'manifest.yml')

    for filename in filenames:
        candidate = target_dir.joinpath(filename)

        if candidate.is_file():
            return candidate

    raise ValueError(f'No manifest found in {target_dir}')



from typing import Any

from pydantic import BaseModel

from liti.core.context import Context
from liti.core.model.v1.datatype import Array, BigNumeric, Bytes, Float, Int, Numeric, String
from liti.core.model.v1.schema import MaterializedView, Partitioning, Schema, Table, View
from liti.core.observe.observer import Defaulter, Validator


def set_defaults(model: Any, defaulter: Defaulter, context: Context):
    if isinstance(model, BaseModel):
        for field_name in model.__pydantic_fields__.keys():
            field = getattr(model, field_name)
            set_defaults(field, defaulter, context)

    if isinstance(model, Int):
        defaulter.int_defaults(model, context)
    elif isinstance(model, Float):
        defaulter.float_defaults(model, context)
    elif isinstance(model, Numeric):
        defaulter.numeric_defaults(model, context)
    elif isinstance(model, BigNumeric):
        defaulter.big_numeric_defaults(model, context)
    elif isinstance(model, Partitioning):
        defaulter.partitioning_defaults(model, context)
    elif isinstance(model, Table):
        defaulter.table_defaults(model, context)
    elif isinstance(model, View):
        defaulter.view_defaults(model, context)
    elif isinstance(model, MaterializedView):
        defaulter.materialized_view_defaults(model, context)
    elif isinstance(model, tuple | list | set):
        for item in model:
            set_defaults(item, defaulter, context)
    elif isinstance(model, dict):
        for item in model.values():
            set_defaults(item, defaulter, context)


def validate_model(model: Any, validator: Validator, context: Context):
    if isinstance(model, BaseModel):
        for field_name in model.__pydantic_fields__.keys():
            field = getattr(model, field_name)
            validate_model(field, validator, context)

    if isinstance(model, Schema):
        validator.validate_schema(model, context)
    elif isinstance(model, Int):
        validator.validate_int(model, context)
    elif isinstance(model, Float):
        validator.validate_float(model, context)
    elif isinstance(model, Numeric):
        validator.validate_numeric(model, context)
    elif isinstance(model, BigNumeric):
        validator.validate_big_numeric(model, context)
    elif isinstance(model, String):
        validator.validate_string(model, context)
    elif isinstance(model, Bytes):
        validator.validate_bytes(model, context)
    elif isinstance(model, Array):
        validator.validate_array(model, context)
    elif isinstance(model, Partitioning):
        validator.validate_partitioning(model, context)
    elif isinstance(model, View):
        validator.validate_view(model, context)
    elif isinstance(model, MaterializedView):
        validator.validate_materialized_view(model, context)
    elif isinstance(model, tuple | list | set):
        for item in model:
            validate_model(item, validator, context)
    elif isinstance(model, dict):
        for item in model.values():
            validate_model(item, validator, context)



from liti.core.context import Context
from liti.core.model.v1.operation.data import table as d
from liti.core.model.v1.operation.ops.base import OperationOps


class CreateSchemaOps(OperationOps):
    op: d.CreateSchema

    def __init__(self, op: d.CreateSchema, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.create_schema(self.op.schema_object)

    def down(self) -> d.DropSchema:
        return d.DropSchema(schema_name=self.op.schema_object.name)

    def is_up(self) -> bool:
        return self.db_backend.has_schema(self.op.schema_object.name)


class DropSchemaOps(OperationOps):
    op: d.DropSchema

    def __init__(self, op: d.DropSchema, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.drop_schema(self.op.schema_name)

    # TODO: figure out how to recreate the resources within the schema when it was dropped
    def down(self) -> d.CreateSchema:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.CreateSchema(schema_object=sim_schema)

    def is_up(self) -> bool:
        return not self.db_backend.has_schema(self.op.schema_name)


class SetDefaultTableExpirationOps(OperationOps):
    op: d.SetDefaultTableExpiration

    def __init__(self, op: d.SetDefaultTableExpiration, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_default_table_expiration(self.op.schema_name, self.op.expiration)

    def down(self) -> d.SetDefaultTableExpiration:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)

        return d.SetDefaultTableExpiration(
            schema_name=self.op.schema_name,
            expiration=sim_schema.default_table_expiration,
        )

    def is_up(self) -> bool:
        return self.db_backend.get_schema(self.op.schema_name).default_table_expiration == self.op.expiration


class SetDefaultPartitionExpirationOps(OperationOps):
    op: d.SetDefaultPartitionExpiration

    def __init__(self, op: d.SetDefaultPartitionExpiration, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_default_partition_expiration(self.op.schema_name, self.op.expiration)

    def down(self) -> d.SetDefaultPartitionExpiration:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)

        return d.SetDefaultPartitionExpiration(
            schema_name=self.op.schema_name,
            expiration=sim_schema.default_partition_expiration,
        )

    def is_up(self) -> bool:
        return self.db_backend.get_schema(self.op.schema_name).default_partition_expiration == self.op.expiration


class SetDefaultKmsKeyNameOps(OperationOps):
    op: d.SetDefaultKmsKeyName

    def __init__(self, op: d.SetDefaultKmsKeyName, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_default_kms_key_name(self.op.schema_name, self.op.key_name)

    def down(self) -> d.SetDefaultKmsKeyName:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetDefaultKmsKeyName(schema_name=self.op.schema_name, key_name=sim_schema.default_kms_key_name)

    def is_up(self) -> bool:
        return self.db_backend.get_schema(self.op.schema_name).default_kms_key_name == self.op.key_name


class SetFailoverReservationOps(OperationOps):
    op: d.SetFailoverReservation

    def __init__(self, op: d.SetFailoverReservation, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_failover_reservation(self.op.schema_name, self.op.reservation)

    def down(self) -> d.SetFailoverReservation:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetFailoverReservation(schema_name=self.op.schema_name, reservation=sim_schema.failover_reservation)

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not provide the failover reservation, always apply
            return False
        else:
            return self.db_backend.get_schema(self.op.schema_name).failover_reservation == self.op.reservation


class SetCaseSensitiveOps(OperationOps):
    op: d.SetCaseSensitive

    def __init__(self, op: d.SetCaseSensitive, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_case_sensitive(self.op.schema_name, self.op.case_sensitive)

    def down(self) -> d.SetCaseSensitive:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetCaseSensitive(schema_name=self.op.schema_name, case_sensitive=sim_schema.is_case_sensitive)

    def is_up(self) -> bool:
        return self.db_backend.get_schema(self.op.schema_name).is_case_sensitive is self.op.case_sensitive


class SetIsPrimaryReplicaOps(OperationOps):
    op: d.SetIsPrimaryReplica

    def __init__(self, op: d.SetIsPrimaryReplica, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_is_primary_replica(self.op.schema_name, self.op.is_primary)

    def down(self) -> d.SetIsPrimaryReplica:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetIsPrimaryReplica(schema_name=self.op.schema_name, is_primary=sim_schema.is_primary_replica)

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not say if it is the primary replica, always apply
            return False
        else:
            return self.db_backend.get_schema(self.op.schema_name).is_primary_replica is self.op.is_primary


class SetPrimaryReplicaOps(OperationOps):
    op: d.SetPrimaryReplica

    def __init__(self, op: d.SetPrimaryReplica, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_primary_replica(self.op.schema_name, self.op.replica)

    def down(self) -> d.SetPrimaryReplica:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetPrimaryReplica(schema_name=self.op.schema_name, replica=sim_schema.primary_replica)

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not provide the primary replica, always apply
            return False
        else:
            return self.db_backend.get_schema(self.op.schema_name).primary_replica == self.op.replica


class SetMaxTimeTravelOps(OperationOps):
    op: d.SetMaxTimeTravel

    def __init__(self, op: d.SetMaxTimeTravel, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_max_time_travel(self.op.schema_name, self.op.duration)

    def down(self) -> d.SetMaxTimeTravel:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetMaxTimeTravel(schema_name=self.op.schema_name, duration=sim_schema.max_time_travel)

    def is_up(self) -> bool:
        return self.db_backend.get_schema(self.op.schema_name).max_time_travel == self.op.duration


class SetStorageBillingOps(OperationOps):
    op: d.SetStorageBilling

    def __init__(self, op: d.SetStorageBilling, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_storage_billing(self.op.schema_name, self.op.storage_billing)

    def down(self) -> d.SetStorageBilling:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_schema = sim_db.get_schema(self.op.schema_name)
        return d.SetStorageBilling(schema_name=self.op.schema_name, storage_billing=sim_schema.storage_billing)

    def is_up(self) -> bool:
        return self.db_backend.get_schema(self.op.schema_name).storage_billing == self.op.storage_billing


class CreateTableOps(OperationOps):
    op: d.CreateTable

    def __init__(self, op: d.CreateTable, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.create_table(self.op.table)

    def down(self) -> d.DropTable:
        return d.DropTable(table_name=self.op.table.name)

    def is_up(self) -> bool:
        return self.db_backend.has_table(self.op.table.name)


class DropTableOps(OperationOps):
    op: d.DropTable

    def __init__(self, op: d.DropTable, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.drop_table(self.op.table_name)

    def down(self) -> d.CreateTable:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return d.CreateTable(table=sim_table)

    def is_up(self) -> bool:
        return not self.db_backend.has_table(self.op.table_name)


class RenameTableOps(OperationOps):
    op: d.RenameTable

    def __init__(self, op: d.RenameTable, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.rename_table(self.op.from_name, self.op.to_name)

    def down(self) -> d.RenameTable:
        return d.RenameTable(
            from_name=self.op.from_name.with_name(self.op.to_name),
            to_name=self.op.from_name.name,
        )

    def is_up(self) -> bool:
        return self.db_backend.has_table(self.op.from_name.with_name(self.op.to_name))


class SetPrimaryKeyOps(OperationOps):
    op: d.SetPrimaryKey

    def __init__(self, op: d.SetPrimaryKey, context: Context):
        self.op = op
        self.context = context

    def up(self):
        # TODO: implement removal of the existing primary key if any
        self.db_backend.set_primary_key(self.op.table_name, self.op.primary_key)

    def down(self) -> d.SetPrimaryKey:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return d.SetPrimaryKey(table_name=self.op.table_name, primary_key=sim_table.primary_key)

    def is_up(self) -> bool:
        return self.db_backend.get_table(self.op.table_name).primary_key == self.op.primary_key


class AddForeignKeyOps(OperationOps):
    op: d.AddForeignKey

    def __init__(self, op: d.AddForeignKey, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.add_foreign_key(self.op.table_name, self.op.foreign_key)

    def down(self) -> d.DropConstraint:
        return d.DropConstraint(table_name=self.op.table_name, constraint_name=self.op.foreign_key.name)

    def is_up(self) -> bool:
        fk = self.op.foreign_key
        return self.db_backend.get_table(self.op.table_name).foreign_key_map.get(fk.name) == fk


class DropConstraintOps(OperationOps):
    op: d.DropConstraint

    def __init__(self, op: d.DropConstraint, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.drop_constraint(self.op.table_name, self.op.constraint_name)

    def down(self) -> d.AddForeignKey:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return d.AddForeignKey(
            table_name=self.op.table_name,
            foreign_key=sim_table.foreign_key_map[self.op.constraint_name],
        )

    def is_up(self) -> bool:
        return self.op.constraint_name not in self.db_backend.get_table(self.op.table_name).foreign_key_map


class SetPartitionExpirationOps(OperationOps):
    op: d.SetPartitionExpiration

    def __init__(self, op: d.SetPartitionExpiration, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_partition_expiration(self.op.table_name, self.op.expiration)

    def down(self) -> d.SetPartitionExpiration:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return d.SetPartitionExpiration(
            table_name=self.op.table_name,
            expiration=sim_table.partitioning and sim_table.partitioning.expiration,
        )

    def is_up(self) -> bool:
        partitioning = self.db_backend.get_table(self.op.table_name).partitioning
        return (partitioning and partitioning.expiration) == self.op.expiration


class SetRequirePartitionFilterOps(OperationOps):
    op: d.SetRequirePartitionFilter

    def __init__(self, op: d.SetRequirePartitionFilter, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_require_partition_filter(self.op.table_name, self.op.require_filter)

    def down(self) -> d.SetRequirePartitionFilter:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return d.SetRequirePartitionFilter(
            table_name=self.op.table_name,
            require_filter=sim_table.partitioning and sim_table.partitioning.require_filter,
        )

    def is_up(self) -> bool:
        partitioning = self.db_backend.get_table(self.op.table_name).partitioning
        return (partitioning and partitioning.require_filter) == self.op.require_filter


class SetClusteringOps(OperationOps):
    op: d.SetClustering

    def __init__(self, op: d.SetClustering, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_clustering(self.op.table_name, self.op.column_names)

    def down(self) -> d.SetClustering:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return d.SetClustering(table_name=self.op.table_name, column_names=sim_table.clustering)

    def is_up(self) -> bool:
        return self.db_backend.get_table(self.op.table_name).clustering == self.op.column_names


class SetFriendlyNameOps(OperationOps):
    op: d.SetFriendlyName

    def __init__(self, op: d.SetFriendlyName, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_friendly_name(self.op.entity_name, self.op.friendly_name)

    def down(self) -> d.SetFriendlyName:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetFriendlyName(entity_name=self.op.entity_name, friendly_name=sim_entity.friendly_name)
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        entity = self.get_entity(self.op.entity_name)
        return entity is not None and entity.friendly_name == self.op.friendly_name


class SetDescriptionOps(OperationOps):
    op: d.SetDescription

    def __init__(self, op: d.SetDescription, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_description(self.op.entity_name, self.op.description)

    def down(self) -> d.SetDescription:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetDescription(entity_name=self.op.entity_name, description=sim_entity.description)
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        entity = self.get_entity(self.op.entity_name)
        return entity is not None and entity.description == self.op.description


class SetLabelsOps(OperationOps):
    op: d.SetLabels

    def __init__(self, op: d.SetLabels, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_labels(self.op.entity_name, self.op.labels)

    def down(self) -> d.SetLabels:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetLabels(entity_name=self.op.entity_name, labels=sim_entity.labels)
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        entity = self.get_entity(self.op.entity_name)
        return entity is not None and entity.labels == self.op.labels


class SetTagsOps(OperationOps):
    op: d.SetTags

    def __init__(self, op: d.SetTags, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_tags(self.op.entity_name, self.op.tags)

    def down(self) -> d.SetTags:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetTags(entity_name=self.op.entity_name, tags=sim_entity.tags)
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        entity = self.get_entity(self.op.entity_name)
        return entity is not None and entity.tags == self.op.tags


class SetExpirationTimestampOps(OperationOps):
    op: d.SetExpirationTimestamp

    def __init__(self, op: d.SetExpirationTimestamp, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_expiration_timestamp(self.op.entity_name, self.op.expiration_timestamp)

    def down(self) -> d.SetExpirationTimestamp:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetExpirationTimestamp(
                entity_name=self.op.entity_name,
                expiration_timestamp=sim_entity.expiration_timestamp,
            )
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        entity = self.get_entity(self.op.entity_name)
        return entity is not None and entity.expiration_timestamp == self.op.expiration_timestamp


class SetDefaultRoundingModeOps(OperationOps):
    op: d.SetDefaultRoundingMode

    def __init__(self, op: d.SetDefaultRoundingMode, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_default_rounding_mode(self.op.entity_name, self.op.rounding_mode)

    def down(self) -> d.SetDefaultRoundingMode:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetDefaultRoundingMode(
                entity_name=self.op.entity_name,
                rounding_mode=sim_entity.default_rounding_mode,
            )
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not provide the default rounding mode, always apply
            return False
        else:
            entity = self.get_entity(self.op.entity_name)
            return entity is not None and entity.default_rounding_mode == self.op.rounding_mode


class SetMaxStalenessOps(OperationOps):
    op: d.SetMaxStaleness

    def __init__(self, op: d.SetMaxStaleness, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_max_staleness(self.op.entity_name, self.op.max_staleness)

    def down(self) -> d.SetMaxStaleness:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_entity = self.get_entity(self.op.entity_name, sim_db)

        if sim_entity is not None:
            return d.SetMaxStaleness(entity_name=self.op.entity_name, max_staleness=sim_entity.max_staleness)
        else:
            raise ValueError(f'Entity {self.op.entity_name} not found in simulated database')

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API provides max staleness in a custom string format that needs to be parsed, always apply
            # TODO: parse the big query value and use it here
            return False
        else:
            entity = self.get_entity(self.op.entity_name)
            return entity is not None and entity.max_staleness == self.op.max_staleness


class SetEnableChangeHistoryOps(OperationOps):
    op: d.SetEnableChangeHistory

    def __init__(self, op: d.SetEnableChangeHistory, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_enable_change_history(self.op.table_name, self.op.enabled)

    def down(self) -> d.SetEnableChangeHistory:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return d.SetEnableChangeHistory(table_name=self.op.table_name, enabled=sim_table.enable_change_history)

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not provide the change history flag, always apply
            return False
        else:
            return self.db_backend.get_table(self.op.table_name).enable_change_history == self.op.enabled


class SetEnableFineGrainedMutationsOps(OperationOps):
    op: d.SetEnableFineGrainedMutations

    def __init__(self, op: d.SetEnableFineGrainedMutations, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_enable_fine_grained_mutations(self.op.table_name, self.op.enabled)

    def down(self) -> d.SetEnableFineGrainedMutations:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return d.SetEnableFineGrainedMutations(
            table_name=self.op.table_name,
            enabled=sim_table.enable_fine_grained_mutations,
        )

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not provide the fine grained mutations flag, always apply
            return False
        else:
            return self.db_backend.get_table(self.op.table_name).enable_fine_grained_mutations == self.op.enabled


class SetKmsKeyNameOps(OperationOps):
    op: d.SetKmsKeyName

    def __init__(self, op: d.SetKmsKeyName, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.db_backend.set_kms_key_name(self.op.table_name, self.op.key_name)

    def down(self) -> d.SetKmsKeyName:
        sim_db = self.simulate(self.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return d.SetKmsKeyName(table_name=self.op.table_name, key_name=sim_table.kms_key_name)

    def is_up(self) -> bool:
        # circular import
        from liti.core.backend.bigquery import BigQueryDbBackend

        if isinstance(self.db_backend, BigQueryDbBackend):
            # the big query API does not provide the kms key name, always apply
            return False
        else:
            return self.db_backend.get_table(self.op.table_name).kms_key_name == self.op.key_name



# A collection of re-imports to allow for convenient namespacing of bigquery types like so:
# from liti import bigquery as bq

# noinspection PyUnresolvedReferences
from google.cloud.bigquery import (
    Client,
    ConnectionProperty,
    Dataset,
    DatasetReference,
    PartitionRange,
    QueryJob,
    QueryJobConfig,
    RangePartitioning,
    ScalarQueryParameter,
    SchemaField,
    Table,
    TableReference,
)

# noinspection PyUnresolvedReferences
from google.cloud.bigquery.enums import RoundingMode
# noinspection PyUnresolvedReferences
from google.cloud.bigquery.schema import _DEFAULT_VALUE as SCHEMA_DEFAULT_VALUE

# noinspection PyUnresolvedReferences
from google.cloud.bigquery.table import (
    BigLakeConfiguration,
    ColumnReference,
    ForeignKey,
    PrimaryKey,
    RowIterator,
    TableConstraints,
    TableListItem,
    TimePartitioning,
)



from typing import Any, ClassVar

from pydantic import field_serializer, field_validator
from pydantic_core.core_schema import FieldSerializationInfo

from liti.core.model.v1.datatype import Datatype, parse_datatype
from liti.core.model.v1.operation.data.base import EntityKind, Operation
from liti.core.model.v1.schema import Column, ColumnName, FieldPath, QualifiedName, RoundingMode


class AddColumn(Operation):
    table_name: QualifiedName
    column: Column

    KIND: ClassVar[str] = 'add_column'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class DropColumn(Operation):
    table_name: QualifiedName
    column_name: ColumnName

    KIND: ClassVar[str] = 'drop_column'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class RenameColumn(Operation):
    table_name: QualifiedName
    from_name: ColumnName
    to_name: ColumnName

    KIND: ClassVar[str] = 'rename_column'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetColumnDatatype(Operation):
    table_name: QualifiedName
    column_name: ColumnName
    datatype: Datatype

    KIND: ClassVar[str] = 'set_column_datatype'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}

    @field_validator('datatype', mode='before')
    @classmethod
    def validate_datatype(cls, value: Datatype | str | dict[str, Any]) -> Datatype:
        return parse_datatype(value)

    @field_serializer('datatype')
    @classmethod
    def serialize_datatype(cls, value: Datatype, info: FieldSerializationInfo) -> str | dict[str, Any]:
        # necessary to call the subclass serializer, otherwise pydantic uses Datatype
        return value.model_dump(exclude_none=info.exclude_none)


class AddColumnField(Operation):
    table_name: QualifiedName
    field_path: FieldPath
    datatype: Datatype

    KIND: ClassVar[str] = 'add_column_field'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}

    @field_validator('datatype', mode='before')
    @classmethod
    def validate_datatype(cls, value: Datatype | str | dict[str, Any]) -> Datatype:
        return parse_datatype(value)

    @field_serializer('datatype')
    @classmethod
    def serialize_datatype(cls, value: Datatype, info: FieldSerializationInfo) -> str | dict[str, Any]:
        # necessary to call the subclass serializer, otherwise pydantic uses Datatype
        return value.model_dump(exclude_none=info.exclude_none)


class DropColumnField(Operation):
    table_name: QualifiedName
    field_path: FieldPath

    KIND: ClassVar[str] = 'drop_column_field'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetColumnNullable(Operation):
    table_name: QualifiedName
    column_name: ColumnName
    nullable: bool

    KIND: ClassVar[str] = 'set_column_nullable'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetColumnDescription(Operation):
    table_name: QualifiedName
    column_name: ColumnName
    description: str | None = None

    KIND: ClassVar[str] = 'set_column_description'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}


class SetColumnRoundingMode(Operation):
    table_name: QualifiedName
    column_name: ColumnName
    rounding_mode: RoundingMode | None = None

    KIND: ClassVar[str] = 'set_column_rounding_mode'

    @property
    def supported_entity_kinds(self) -> set[EntityKind]:
        return {'TABLE'}



import json
import logging
from pathlib import Path
from typing import Literal

import yaml
from devtools import pformat

from liti.core.backend.base import DbBackend, MetaBackend
from liti.core.context import Context
from liti.core.file import get_manifest_path
from liti.core.function import attach_ops
from liti.core.logger import NoOpLogger
from liti.core.model.v1.manifest import Manifest
from liti.core.model.v1.operation.data.base import Operation
from liti.core.model.v1.operation.data.table import CreateSchema, CreateTable
from liti.core.model.v1.parse import parse_manifest, parse_operations, parse_templates
from liti.core.model.v1.schema import DatabaseName, Identifier, QualifiedName, SchemaName
from liti.core.model.v1.template import Template
from liti.core.observe import set_defaults, validate_model

log = logging.getLogger(__name__)


def apply_templates(file_operations: list[tuple[Path, list[Operation]]], templates: list[Template]):
    # first collect all the update functions
    update_fns = [
        # setting default values to work around late binding pass-by-reference closures
        lambda fn=update_fn, v=template.value: fn(v)
        for filename, ops in file_operations
        for template in templates
        if not template.files or filename in {Path(f) for f in template.files}
        for op in ops
        if not template.operation_kinds or type(op) in [Operation.by_kind(k) for k in template.operation_kinds]
        for root, root_match in (
            op.get_roots(template.root_type, template.full_match)
            if template.root_type is not None
            else [(op, template.full_match)]
        )
        for update_fn in root.get_update_fns(template.path, [template.local_match, root_match])
    ]

    # Then apply them so templates do not read the replacements of other templates.
    # If multiple templates update the same field, the last one wins.
    for fn in update_fns:
        fn()


class MigrateRunner:
    def __init__(self, context: Context):
        self.context = context

    @property
    def db_backend(self) -> DbBackend:
        return self.context.db_backend

    @property
    def meta_backend(self) -> MetaBackend:
        return self.context.meta_backend

    @property
    def target_dir(self) -> Path | None:
        return self.context.target_dir

    @property
    def manifest(self) -> Manifest:
        if self.context.manifest is None:
            self.context.manifest = parse_manifest(get_manifest_path(self.target_dir))

        return self.context.manifest

    @property
    def templates(self) -> list[Template] | None:
        if self.context.templates is None and self.context.template_files:
            self.context.templates = [
                template
                for path in self.context.template_files
                for template in parse_templates(path)
            ]

        return self.context.templates

    @property
    def target_operations(self) -> list[Operation]:
        if self.context.target_operations is None:
            manifest = self.manifest
            file_operations = parse_operations(manifest.operation_files, self.target_dir)

            if self.templates:
                apply_templates(file_operations, self.templates)

            self.context.target_operations = [op for _, ops in file_operations for op in ops]

        return self.context.target_operations

    def run(
        self,
        wet_run: bool | None = None,
        allow_down: bool | None = None,
    ):
        """
        :param wet_run: [False] True to run the migrations, False to simulate them
        :param allow_down: [False] True to allow down migrations, False will raise if down migrations are required
        """

        wet_run = wet_run if wet_run is not None else False
        allow_down = allow_down if allow_down is not None else False
        logger = NoOpLogger() if self.context.silent else log

        for op in self.target_operations:
            set_defaults(op, self.db_backend, self.context)
            validate_model(op, self.db_backend, self.context)

        if wet_run:
            self.meta_backend.initialize()

        migration_plan = self.meta_backend.get_migration_plan(self.target_operations)

        if not allow_down and migration_plan['down']:
            raise RuntimeError('Down migrations required but not allowed. Use --down')

        def apply_operations(operations: list[Operation], up: bool):
            for op in operations:
                if up:
                    up_op = op
                else:
                    # Down migrations apply the inverse operation
                    up_op = attach_ops(op, self.context).down()

                up_ops = attach_ops(up_op, self.context)

                # Apply only if not applied already
                if not up_ops.is_up():
                    logger.info(pformat(up_op, highlight=True))

                    if wet_run:
                        up_ops.up()

                # Update the metadata
                if wet_run:
                    if up:
                        self.meta_backend.apply_operation(op)
                    else:
                        self.meta_backend.unapply_operation(op)

        logger.info('Down')
        apply_operations(migration_plan['down'], False)
        logger.info('Up')
        apply_operations(migration_plan['up'], True)
        logger.info('Done')


def sort_operations(operations: list[Operation]) -> list[Operation]:
    """ Sorts the operations into a valid application order """
    create_schemas: list[CreateSchema] = []
    create_tables: dict[QualifiedName, CreateTable] = {}
    others: list[Operation] = []

    for op in operations:
        if isinstance(op, CreateSchema):
            create_schemas.append(op)
        elif isinstance(op, CreateTable):
            create_tables[op.table.name] = op
        else:
            others.append(op)

    sorted_ops = {}

    while create_tables:
        satisfied_ops: dict[QualifiedName, CreateTable] = {}

        for op in create_tables.values():
            if all(fk.foreign_table_name in sorted_ops for fk in op.table.foreign_keys or []):
                satisfied_ops[op.table.name] = op

        if not satisfied_ops:
            raise RuntimeError('Unsatisfied or circular foreign key references found')

        sorted_ops.update(satisfied_ops)

        for table_name in satisfied_ops:
            del create_tables[table_name]

    return create_schemas + list(sorted_ops.values()) + others


class ScanRunner:
    def __init__(self, context: Context):
        self.context = context

    @property
    def db_backend(self) -> DbBackend:
        return self.context.db_backend

    def run(
        self,
        database: DatabaseName,
        schema: SchemaName,
        table: Identifier | None = None,
        format: Literal['json', 'yaml'] = 'yaml',
    ):
        """
        :param database: database to scan
        :param schema: schema to scan
        :param table: [None] None scans the whole schema, otherwise scans only the provided table
        :param format: ['yaml'] the format to use when printing the operations
        """

        validate_model(database, self.db_backend, self.context)
        validate_model(schema, self.db_backend, self.context)

        if table:
            validate_model(table, self.db_backend, self.context)

            table_name = QualifiedName(database=database, schema=schema, name=table)
            create_table = self.db_backend.scan_relation(table_name)

            if create_table is None:
                raise RuntimeError(f'Table not found: {table_name}')

            operations = [create_table]
        else:
            operations = sort_operations(self.db_backend.scan_schema(database, schema))

        op_data = [op.to_op_data(format=format) for op in operations]

        file_data = {
            'version': 1,
            'operations': op_data,
        }

        if format == 'json':
            print(json.dumps(file_data, indent=4, sort_keys=False))
        elif format == 'yaml':
            print(yaml.dump(file_data, indent=2, sort_keys=False))
        else:
            raise ValueError(f'Unsupported format: {format}')



from typing import Any, Iterator

from liti.core.context import Context
from liti.core.model.v1.datatype import Array, Datatype, Struct
from liti.core.model.v1.operation.data.base import Operation
from liti.core.model.v1.operation.ops.base import OperationOps
from liti.core.model.v1.schema import ColumnName, FieldPath, Table


def extract_nested(data: Any, iterator: Iterator[Any], get_next: callable) -> Any:
    def extract(data: Any) -> Any:
        try:
            item = next(iterator)
        except StopIteration:
            return data

        return extract(get_next(data, item))

    return extract(data)


def extract_nested_datatype(table: Table, field_path: FieldPath) -> Datatype:
    def get_next(dt: Datatype, segment: str) -> Datatype:
        if isinstance(dt, Array):
            return get_next(dt.inner, segment)
        elif isinstance(dt, Struct):
            if segment in dt.fields:
                return dt.fields[segment]

        raise ValueError(f'Unable to extract datatype from {table} with path {field_path}')

    segments = iter(field_path)
    datatype = table.column_map[ColumnName(next(segments))].datatype
    return extract_nested(datatype, segments, get_next)


def attach_ops(operation: Operation, context: Context) -> OperationOps:
    return OperationOps.get_attachment(operation)(operation, context)
